from dataclasses import dataclass, field
from astropy.constants.astropyconst40 import M_sun
from divergent_gravity.classes.bodys.body import Body
import scipy.constants as spc


LANIAKEA_RADIUS = 250_000_000 * spc.light_year


@dataclass
class Laniakea(Body):
    mass: float = field(default_factory=lambda: M_sun.value * 10**17)
    radius: float = field(default_factory=lambda: LANIAKEA_RADIUS)
